# -*- coding:utf-8 -*-
import datetime
import json
import kafka
import time
from proto import canal_packet_pb2 as packet_pb2
from proto import canal_entry_pb2 as entry_pb2

PARTITION_SEQ_KEY = "ps"
PARTITION_SEQ2_KEY = "ps2"
CHECKPOINT_KEY = "Checkpoint"
HEARTBEAT_KEY = "Heartbeat"


## 本Demo基本执行流程为：
# 1.创建 Kafka 消费者，不停消费消息；
# 2.依次遍历每条消息，根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理；
# 3. partitionMsgConsumer 将原始消息反序列化为 Entry 或者 FlatMessage 结构，传递到 onEntry 方法然后做进一步的处理。
#    onEntry 方法里本Demo仅对数据做打印处理，您可以根据业务需要修改这里的逻辑。
def block_consume(brokers, topic, group, user, password, is_flat):
    # create a kafka consumer
    if user != "":
        consumer = kafka.KafkaConsumer(topic,
                                       auto_offset_reset='earliest',
                                       enable_auto_commit=False,
                                       group_id=group,
                                       sasl_plain_username=user,
                                       sasl_plain_password=password,
                                       sasl_mechanism='SCRAM-SHA-512',
                                       security_protocol='SASL_PLAINTEXT',
                                       bootstrap_servers=brokers)
    else:
        consumer = kafka.KafkaConsumer(topic,
                                       auto_offset_reset='earliest',
                                       enable_auto_commit=False,
                                       group_id=group,
                                       bootstrap_servers=brokers)

    # get partition of topic
    partitions = consumer.partitions_for_topic(topic)
    partition_consumers = {}

    for p in partitions:
        partition_consumers[p] = PartitionMsgConsumer()

    for message in consumer:
        partition_consumers[message.partition].consume(consumer, message, is_flat)


def get_partition_seq(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ_KEY:
            return int(m[1].decode())


def get_partition_seq2(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ2_KEY:
            return int(m[1].decode())
    return -1


def isCheckpoint(message):
    for m in message.headers:
        if m[0] == CHECKPOINT_KEY:
            return True
    return False


def isHeartbeat(message):
    for m in message.headers:
        if m[0] == HEARTBEAT_KEY:
            return True
    return False


def timestamp2str(timestamp):
    date_time = datetime.fromtimestamp(timestamp)
    return date_time.strftime("%Y-%m-%d %H:%M:%S")


#  您可以在此函数基础上开发满足自己需要的数据处理逻辑。
def on_entry(partition, offset, ps, msg_time, entry, flat_msg):
    if flat_msg is not None:
        print(
            "-->[partition: {}, offset: {}, partitionSeq: {}, msgSentTime: {}, currentTime: {}]\n[{}]\n".format(
                partition, offset, ps, msg_time, int(time.time() * 1000), flat_msg)
        )

    if entry is not None:
        header = entry.header
        entry_type = entry.entryType
        if entry_type == entry_pb2.CHECKPOINT:
            cpk = entry_pb2.Checkpoint()
            cpk.ParseFromString(entry.storeValue)
            print(
                "-->[partition: {}, offset: {}, partitionSeq: {}, msgSentTime: {}, currentTime: {}]\n[{}]\n[{}]\n".format(
                    partition, offset, ps, msg_time, int(time.time() * 1000), header, cpk))
        elif entry_type == entry_pb2.TRANSACTIONBEGIN:
            begin = entry_pb2.TransactionBegin()
            begin.ParseFromString(entry.storeValue)
            print(
                "-->[partition: {}, offset: {}, partitionSeq: {}, msgSentTime: {}, currentTime: {}]\n[{}]\n[{}]\n".format(
                    partition, offset, ps, msg_time, int(time.time() * 1000), header, begin))
        elif entry_type == entry_pb2.TRANSACTIONEND:
            end = entry_pb2.TransactionEnd()
            end.ParseFromString(entry.storeValue)
            print(
                "-->[partition: {}, offset: {}, partitionSeq: {}, msgSentTime: {}, currentTime: {}]\n[{}]\n[{}]\n".format(
                    partition, offset, ps, msg_time, int(time.time() * 1000), header, end))
        elif entry_type == entry_pb2.ROWDATA:
            row_change = entry_pb2.RowChange()
            row_change.ParseFromString(entry.storeValue)
            print(
                "-->[partition: {}, offset: {}, partitionSeq: {}, msgSentTime: {}, currentTime: {}]\n[{}]\n[{}]\n".format(
                    partition, offset, ps, msg_time, int(time.time() * 1000), header, row_change))
        else:
            print("ERROR: UnExpected EntryType: {}".format(entry_type.name))
            raise ValueError("ERROR: UnExpected EntryType: {}".format(entry_type.name))


class PartitionMsgConsumer(object):
    def __init__(self):
        self.last_partition_seq = 0
        self.last_partition_seq2 = 0
        self.last_commit_time = 0

    def consume(self, consumer, message, is_flat):
        # 分区内消息的连续编号，当前版本没有用到，忽略
        ps = get_partition_seq2(message)

        if is_flat:  # canal-json message
            try:
                flat_msg = json.loads(message.value)
            except Exception as e:
                return e
            errmsg = on_entry(message.partition, message.offset, ps, message.timestamp, None, flat_msg)
            if errmsg is not None:
                return errmsg
        else:
            # 1、将kafka消息的Value值反序列化为Packet
            packet = packet_pb2.Packet()
            try:
                packet.ParseFromString(message.value)
            except Exception as e:
                print(e)
                return e
            if packet.type != packet_pb2.MESSAGES:
                print("UnExpected PacketType: " + packet.type)
                return None
            # 2、将packet.body反序列化后的Messages的Messages[0]反序列化为Entry
            msgs = packet_pb2.Messages()
            try:
                msgs.ParseFromString(packet.body)
            except Exception as e:
                print(e)
                return e
            entry = entry_pb2.Entry()
            try:
                entry.ParseFromString(msgs.messages[0])
            except Exception as e:
                print(e)
                return e
            # 3、进一步解析处理entry，Demo仅做打印处理，请按照业务需要自行修改
            errmsg = on_entry(message.partition, message.offset, ps, message.timestamp, entry, None)
            if errmsg is not None:
                return errmsg

        # 以下两种Kafka提交位点方式默认同时生效，如有特殊需要可以自行修改 #
        offsets = {
            kafka.TopicPartition(message.topic, message.partition): kafka.OffsetAndMetadata(message.offset + 1, None),
        }
        # commit type 1: 遇到checkpoint消息做一次提交
        if isCheckpoint(message):
            consumer.commit(offsets)
        # commit type 2: 每隔10秒做一次提交
        if self.last_commit_time == 0 or (int(time.time()) - self.last_commit_time) >= 10:
            consumer.commit(offsets)
            self.last_commit_time = int(time.time())
